package ConfigFile;

#
#  Simple package to read and write config files.  Config file
#  format like
#
#  NAME=VALUE # comment
#  # Comment
#
#  blanks ignored
#
#  $Id: ConfigFile.pm,v 1.1 2001/04/04 00:01:46 mschube Exp $
#

use strict;
use Fcntl qw(:flock);

sub LOWERCASE {1;}
sub UPPERCASE {2;}

sub ANY {1;}
sub NAMED {0;}

sub read {
  my $file = shift || die "required parameter filename missing!\n";
  my $cf   = shift;
  my $case = shift;
  my $type = shift;  #  ANY or NAMED

  unless ( $cf && ref($cf) eq 'HASH' ) {
    die "Required HASH reference parameter missing!\n";
  }

  local(*FILE);

  unless ( open(FILE,"< $file") ) {
    die "Can't read from config file $file: $!\n";
  }

  #  Read variables into hash.
  while (<FILE>) {
    next if ( m/^#/o or m/^\s*$/o );
    #  Match NAME=value statement.
    next unless ( m/^\s*(\S+)\s*=s*([^#]+)/ );

    if ( $case == LOWERCASE ) {
        if ( $type == NAMED ) {
            next unless exists $cf->{"\L$1\E"};
        }
    } else {
        if ( $type == NAMED ) {
            next unless exists $cf->{"\U$1\E"};
        }
    }
    
    #  Make key uppercase and add to config hash.
    if ( $case == LOWERCASE ) {
        $cf->{lc($1)} = "$2";
        $cf->{lc($1)} =~ s/\s+$//o;
    } else {
        $cf->{uc($1)} = "$2";
        $cf->{uc($1)} =~ s/\s+$//o;
    }

  }
}

sub write {
  my $file = shift;
  my $cf   = shift;

  unless ( $cf && ref($cf) eq 'HASH' ) {
    die "Required HASH reference parameter missing!\n";
  }

  local(*FILE);

  unless ( open(FILE,"> $file") ) {
    die "Can't write to config file $file: $!\n";
  }

  unless ( flock(FILE, LOCK_EX) ) {
    die "Can't lock $file for exclusive write: $!\n";
  }

  print FILE <<EOF;
#
#  @{[scalar(localtime())]}
#  Config file generated by $0 
#
#  WARNING:  Auto-generated:  Hand edits may be lost!
#
EOF

  local($_);

  for (keys %{$cf}) {
    print FILE "$_=$cf->{$_}\n";
  }

  close(FILE);

  return 1;
}

1;
