package VarSub;

#
#  shellstyle( \$page, \%vars, $sect );
#
#  Substitute values in $vars using keys of %vars hash as variable
#  names and values as values to substitute in text referenced by
#  $text.  if $sect is used, it can be used to keep different
#  sections of macros apart ... i.e.
#
#  substitute_values( \$page, \%bills, 'bill' );
#
#  bill:$bill_id
#
#  Version: 1.11
#

sub shellstyle {
    my $text = shift;  #  Reference to text to do subsitution on.
    my $vref = shift;  #  Reference to hash with variables / values
    my $sect = shift;  #  Section to have different macro domains ...
                       #  i.e. bill:${client:-no client}

    my %vars = %{$vref};

    #  Subsitute special tag $date or ${date} with current date/time
    $$text =~ s/\${?date}?/scalar(localtime(time))/geoi;

    #  Do substitution on content
    #  Vars can be $var, ${var}, or ${var:-NNN}

    local($_);
    for (keys %vars) {

      #  Keep sanity on blank values so we
      #  don't miss cleaning up
      #  $var forms.
      if ( $vars{$_} =~ m#^\s*$#o ) {
        
        #  Substitute vars with blanks if found
        if ( $sect ) {
          $$text =~ s/$sect:\$$_//gi;
        } else {
          $$text =~ s/\$$_//gi;
        }           
        
        #  No need to continue .. no value!
        next;         
      }

      #  Check for ${var:-VAL} form .. since variable
			#  has data, we remove these
      if ( $$text =~ m/\${$_:-/i ) {
        if ( $sect ) {
          $$text =~ s/$sect:\${$_:-[^}]+}//gi;
        } else {
          $$text =~ s/([^:])\${$_:-[^}]+}/$1/gi;
        }
      }
      
      #  If value is not 0 or "" substitute string substitute
      #  specified string
      #  ${var:>value}

      if ( $$text =~ m/\${$_:>/i ) {
        if ( $sect ) {
          $$text =~ s/$sect:\${$_:>([^}]+)}?(\W)/$1$2/gi;
        } else {
          $$text =~ s/([^:])\${$_:>([^}]+)}?(\W)/$1$2$3/gi;
        }
      }
 
      #  if variable is present, subsitute specified vr
      #  like shell ${var:+OK}

      if ( $$text =~ m/\${$_:\+/i ) {
        if ( $sect ) {
          $$text =~ s/$sect:\${$_:\+([^}]+)}?(\W)/$1$2/gi;
        } else {
          $$text =~ s/([^:])\${$_:\+([^}]+)}?(\W)/$1$2$3/gi;
        }
      }

      #  Check for real key/val substitution
      if ( $sect ) {
        $$text =~ s/$sect:\${?$_(?::-[^}]*)?(?:}|(\W))/$vars{$_}$1/gi;
      } else {
        $$text =~ s/\${?$_(?::-[^}]*)?}?(?:}|(\W))/$vars{$_}$1/gi;
      }

    }
    

    #  Substitute defaults in.
    if ( $sect ) {
      $$text =~ s/$sect:\${\w+:-([^}]+)}?(\W)/$1$2/gi;
    } else {
      $$text =~ s/([^:])\${\w+:-([^}]+)}?(\W)/$1$2$3/gi;
    }

    #  Substitute defaults in.
    if ( $sect ) {
      $$text =~ s/$sect:\${\w+:>[^}]+}?(\W)/$1/gi;
    } else {
      $$text =~ s/([^:])\${\w+:>[^}]+}?(\W)/$1$2/gi;
    }

    #  Substitute defaults in.
    if ( $sect ) {
      $$text =~ s/$sect:\${\w+:\+[^}]+}?(\W)/$1/gi;
    } else {
      $$text =~ s/([^:])\${\w+:\+[^}]+}?(\W)/$1$2/gi;
    }

    #  Substitute all left overs with blanks
    if ( $sect ) {
      $$text =~ s/([^:])$sect:?\${?\w+}?/$1/gi;
    } else {
      $$text =~ s/([^:])\${?\w+}?/$1/gi;
    }

    return 1;
}

1;
