#!/usr/local/bin/perl

#  Version $Revision: 1.10 $
#
#  Form processor.  This script will read in a form submitted to it via
#  either GET or POST, process the form, and send out an autoreply,
#  email of the form's output, or both.  All parameters used to configure
#  the scripts' behavior are snt via hidden tags except for the security,
#  logger, and debug options, which are set in the configuration file
#  ($conf variable);
#
#  See INSTRUCTIONS.txt for more info
#
#  This form processor is offered as free software under the same terms as
#  the GNU Public License (GPL).
#
#  Author:  Max Schubert <maxs@webwizarddesign.com>
#

use CGI qw(
    :standard :html3 escape unescape param font center
    remote_host virtual_host
);

use lib './lib';
use LWP::Simple qw(get is_success);    #  get and is_success to get HTML docs 
use Net::SMTP;      #  For form emailing
use Sys::Hostname;  #  For SMTP HELO
use strict;

##################
#  User variable #
##################

#  Config file to use
my $conf = './superfrm.cf';

####################################################################
#  END USER VARS -- All other configuration is through config file #
#  or form tags.                                                   #
####################################################################

#  Release history:
#  See HISTORY

#  Constants for substitue_values routine
sub ERROR_PAGE()  {1;}
sub THANKS_PAGE() {2;}
sub MAIL_FORM()   {3;}
sub AUTO_REPLY()  {4;}

#  Out damn buffer!  You only hold us web folk back!
$| = 1;

#  hash for all config values.
my $cf = {
    's.thanks'    => undef,     #  File to show user after form is submitted
    's.mailto'    => undef,     #  User to email form to.
    's.mailcc'    => undef,     #  Users to CC form output to
    's.mailbcc'   => undef,     #  Users to BCC form output to
    's.mailfrom'  => undef,     #  Who we send mail as.
    's.mailform'  => undef,     #  File to use when emailing form results to 
                                #  mailto person.
    's.subject'   => undef,     #  The subject of our autoreply.
    's.autoreply' => undef,     #  File to use for autoreply.
    's.errorpage' => undef,     #  Page to show when required field is missing
    's.reqfields' => undef,     #  Comma-separated list of required fields.
    's.nomail'    => undef,     #  No email form is sent if this is set
    's.noreply'   => undef,     #  No auto-reply is sent if this is set
    's.formfromsender' => undef,  #  Use address/name set in form as from 
                                  #  address for form output email if set to 1
    check_client    => undef,     #  Set in config file.  If on, allow list is 
                                #  checked when script is called.
    secure        => undef,     #  (old name for check_client .. deprecated)
    allow         => undef,     #  List of sites to allow access to.
    allow_regex   => undef,     #  Regexp to use for denied sites.
    denied_addresses => undef,  #  List of denied IP addresses (clients)
    addr_regex    => undef,     #  Regex of denied IP addresses (clients)
    logging       => undef,     #  if set, logger all uses of script.
    logfile       => 'log.txt', #  If set, use this as log file.
    rootdir       => undef,     #  To have files in chrooted() environment,
                                #  so bad guys can't request /etc/passwd :(
    form       => {},           #  Empty hash for input form stuff
    mailval       => {          #  hash for email values; these vars are
        first     => undef,     #  available for use in templates and are
        'last'    => undef,     #  also used in email_reply().  first, last,
        fullname  => undef,     #  and fullname / name are pulled from 
        email     => undef,     #  the submitting form and have 's.' prepended
        remote_host => undef,   #  to them in the USERS form.
        remote_addr => undef,
        remote_user => undef
    },
    admin_email   => undef,     #  Administrator email address
    email_errors  => undef,     #  Mail errors that occur to the administrator
    mail_to_admin => undef,     #  Bcc admin on autoreply and/or form output 
                                #  for all forms.
    mailhost      => undef,     #  SMTP host
    debug         => undef,     #  Debugging on / off
    debugfile     => undef,     #  File for debugging
    badfields     => undef,     #  Reference to array containing required 
                                #  fields not filled in on received form;
                                #  used by validate_form().

    mf => {                     #  Mail fields we check for; we list these
        's.fullname' => undef,  #  separately because they are used in user's
        's.name'     => undef,  #  forms with out the 's.' and we have to check
        's.email'    => undef,  #  for them specially.
        's.last'     => undef,
        's.first'    => undef
    }
};

#########
#  MAIN #
#########

#  Read logger, debug, and allowed servers from configuration file.  Values
#  are added into $cf hash.
read_config_file($cf,$conf);

#  Make sure the server requesting us is allowed to use us ...
#  also make sure we were not called directly ... we don't want
#  to be called that way.

check_for_valid_user() if ( $cf->{'check_client'} );

#  Log remote user info
logger(
  "$ENV{'HTTP_USER_AGENT'}|$ENV{'HTTP_HOST'}|$ENV{'REMOTE_ADDR'}|" .
  "New connection"
);

#  Read submitted form and put administrative ('s.' variables) and form
#  variables into cf hash.
parse_form($cf);

#  Email the form results to s.mailto if rquested.
format_and_send_form($cf) if ( $cf->{'s.nomail'} == 0 or $cf->{'s.nomail'} eq "" );

#  Send an autoreply to cf->{'mailval'}->{'email'} if requested.
send_reply($cf) if ( $cf->{'s.noreply'} == 0 or $cf->{'s.noreply'} eq "");

#  Output thank you page to form user.
thank_you($cf);

exit 0;

########
# SUBS #
########

#  Open config file and read desired values into $cf hash.

sub read_config_file {
    my ($cf,$conf) = @_;
    #  my ($key,$value);

    local(*CF);

    open (CF,"< $conf") || web_die("Can't read from $conf:  $!");

    #  Read variables into hash.
    while (<CF>) {
        next if ( m/^#/o or m/^\s*$/o );
        #  Match NAME=value statement.
        next unless ( m/^\s*(\S+)\s*=\s*(\S+)(?:\s+#)?/o && 
                      (exists $cf->{"\L$1\E"}) );
        #  Make key lowercase.
        #  ($key,$value) = (lc($1),$2);
        #  Add to config hash.
        debug("\L$1\E has $2");
        $cf->{lc($1)} = "$2";

    }

    close(CF);

    #  If logging is on, make sure we can write to log file.

    if ( $cf->{'logging'} ) {
        $cf->{'logging'} = 1;
        logger("Testing logging system ... 1 2 3");
    }
       
    #  Setup security regexp if needed.  regexp is built from comma
    #  separated value list in allow parameter of config file.
    #  Only used if CHECK_HOST or SECURE (deprecated) variable is set.

    if ( $cf->{'secure'} || $cf->{'check_client'} ) {
        $cf->{'check_client'} = 1;
        foreach ( (split(/\s*,\s*/o, "$cf->{allow}") ) ) {
            s/\./\\\./go;
            $cf->{'allow_regex'} .= "$_\|";
        }
        $cf->{'allow_regex'}  =~ s/\|$//o;
        $cf->{'allow_regex'}  =~ s/\*/\.\*/go;

        #  Do the same for the IP address list
        foreach ( (split(/\s*,\s*/o, "$cf->{denied_addresses}") ) ) {
            s/\./\\\./go;
            $cf->{'addr_regex'} .= "$_\|";
        }
        $cf->{'addr_regex'}  =~ s/\|$//o;
        $cf->{'addr_regex'}  =~ s/\*/\.\*/go;
    }

    debug("allow regex is $cf->{allow_regex}");
    debug("IP regex is $cf->{addr_regex}");

    #  Check if admin email is missing IF email errors or mail to admin
    #  options are used.

    if ( 
        (! $cf->{'admin_email'}) && 
        ($cf->{'email_errors'} || $cf->{'mail_to_admin'}) ) {

        web_die(
            "Administrative email address (ADMIN_EMAIL) is required if " .
            "either EMAIL_ERRORS or MAIL_TO_ADMIN is used."
        );

    }

    #  If admin email is used, check address and parse mail to admin and
    #  email errors options.

    if ( $cf->{'admin_email'} ) {

        if (! check_email_addr($cf->{'admin_email'})) {
            logger("Administrative email address not valid!");
            web_die(
                "Administrative email address invalid; contact " .
                "administrator for help."
            );
        }

        if ( $cf->{'email_errors'} ) { 
            $cf->{'email_errors'} = 1;
        }

        # MAXX -- TBD ... need symbolic constants and parsing
        if ( $cf->{'mail_to_admin'} ) { 
        }
    }

}

#  Parse input form, setting setup values as necessary and populating
#  form member of passed cf hash with input form values.

sub parse_form {
    my $cf = shift;
    #  Read in non-admin values into hash
    read_form_values($cf);
    #  Check for administrative values and take appropriate actions.
    check_admin_values($cf);
    #  Check for required fields if desired.
    validate_form($cf) if ( $cf->{'s.reqfields'} ne "");
    debug("Validation done.");
}

#  Email the results of the form to the person specified in 
#  $cf->{'s.mailto'}, $cf->{'s.mailcc'}, $cf->{'s.mailbcc'}

sub format_and_send_form {

    my $cf = shift;

    #  Check for user-supplied subject and set default if not found.
    my $subject = $cf->{'s.subject'};
    
    if ( ! $subject ) {
        $subject = "Form submitted by $cf->{'form'}->{fullname} " .
                   "from $cf->{'mailval'}->{'remote_addr'}";
    }

    #  Exit if no email address.
    if ( $cf->{'s.mailto'} eq "" ) {
        logger("No email address (S.MAILTO) for S.MAILFORM; no mail sent");
        return;
    }


    #  Set from address header and real address.  Will use
    #  address / name of sender if s.formfromsender is set.

    my $from_header;
    my $from_real;
    if ( $cf->{'s.formfromsender'} ) {
        $from_real   = $cf->{'form'}->{'email'};
        $from_header = q(") . $cf->{'form'}->{'fullname'} . q(" ) .
                       q(<) . $cf->{'form'}->{'email'}    . q(>);
    } else {
        $from_real    = $cf->{'s.mailfrom'};
        $from_header  = '"Super Form Processor" <' . $cf->{'s.mailfrom'} . '>';
    }

    debug("Mail will be sent from $from_header");
    debug("Mail will be to $cf->{'s.mailto'}");

    #  Add CC list if present

    my $cc_list;

    #  Create Cc header list

    if ($cf->{'s.mailcc'}) {
      $cc_list = join(', ', @{$cf->{'s.mailcc'}});
      debug("Mail will be CC'd to $cc_list");
    }

    my $fref;  #  Form text reference

    #  Check to see if we have a template to use ... if so , use it
    if ( $cf->{'s.mailform'} ne "" ) {

        $fref = \$cf->{'s.mailform'};

        debug("Using custom mail form for form email");
        #  Do variable substitution.
        substitute_values($fref, MAIL_FORM);

    } else {

        debug("Using canned mail form for form email");

        #  Our canned reply.
 
        my $body = "\n>> Form data\n";

        #  Add all mailval keys to output.
        for ( keys %{$cf->{'mailval'}} ) {
            $body .= sprintf("%-25s%s",$_,"$cf->{'mailval'}->{$_}\n");
        }

        #  Add all form keys to output
        for ( keys %{$cf->{'form'}} ) {
            $body .= sprintf("%-25s%s",$_,"$cf->{'form'}->{$_}\n");
        }
        #  Add footer.
        $body = join('',$body, ">> Form received at ",  
                        scalar(localtime(time)), "\n");

        $fref = \$body;
    }
        
    #  Send it.
    mail_form(
        $fref, $subject, $from_header, $from_real, $cf->{'s.mailto'}, 
        $cf->{'s.mailcc'}, $cf->{'s.mailbcc'}
    );

}

#  Send a reply to the person submitting the form.  Use the template 
#  in cf->{s.autoreply} if available, else use our own.

sub send_reply {
    my $cf = shift;
    my $tref;   #  For canned reply text.

    #  Exit if no email address.
    if ( $cf->{'mailval'}->{'email'} eq "" ) {
        logger("No email address (S.MAILTO) for S.MAILFORM; no mail sent");
        return;
    }

    #  Check for custom subject
    my $subject = $cf->{'s.subject'};
    $subject = "Your form submission" if ( $subject =~ m/^\s*$/o );

    #  Send an autoreply using custom template or canned one if no
    #  custom template is found.

    if ( $cf->{'s.autoreply'} ne "" ) {

        debug("Custom rply $cf->{'s.mailfrom'} -> $cf->{'mailval'}->{'email'}");
        substitute_values(\$cf->{'s.autoreply'}, AUTO_REPLY);
        
        mail_form(
            \$cf->{'s.autoreply'}, $subject, $cf->{'s.mailfrom'},
            $cf->{'s.mailfrom'}, $cf->{'mailval'}->{'email'}
        );

    } else {
        #  Canned reply.

        debug("Canned rply $cf->{'s.mailfrom'} -> $cf->{'mailval'}->{'email'}");

        $tref = <<EOF;

Dear $cf->{'mailval'}->{'first'},
    This email is being sent to you to confirm that we have received your
form input.  Thank you for taking the time to fill out our form.

Form sent from $cf->{'mailval'}->{'remote_addr'} by $cf->{'mailval'}->{'fullname'}.

Sincerely,
The administrators

EOF
        mail_form(
            \$tref, $subject, $cf->{'s.mailfrom'}, $cf->{'s.mailfrom'},
            $cf->{'mailval'}->{'email'}
        );
    }

}

sub thank_you {
    my $cf = shift;

    #  Send thank you page to user.
    if ( $cf->{'s.thanks'} ne "" ) {
    
        #  Send custom page.
        debug("Using custom thank you page");

        substitute_values(\$cf->{'s.thanks'},THANKS_PAGE);
        print header(), $cf->{'s.thanks'};
        
    } else {

        #  Canned thanks page.
        debug("Using canned thank you page");

        print
        header(),
        start_html( { '-title' => "Thank you for filling out our form" } ),
        center(
            font( 
                { -face => 'Arial' },
                h1("Thank you for filling out our form")
            )
        ),
        hr(),
        font( 
            { -face => 'Arial' },
            qq(
                $cf->{'mailval'}->{'first'}:<BR>
                Thank you for taking the time to fill out our form; we 
                appreciate your input.  &nbsp;Please let the administrator
                of this site know if you have any questions about any of the
                questions you completed.
            )
        ),
        hr(),
        end_html();
    }
}

#  Send HTML-ized error message to web and exit.

#  Flag to keep web_die from recursing with email failures
my $SENT_ERROR = 0;

sub web_die {
    my $msg = shift;

    exit 0 if ( $SENT_ERROR );

    #  Note if we were called by logger to avoid recursive error calls
    my $from_logger = ( (caller(1))[3] eq 'main::logger') ? 1 : 0;

    print header(),
    start_html(
        { '-title' => 'Error occurred while running script' }
    ),
    h1(
        font( 
            { -face => 'Arial' , -size => 4 },
            "Error occurred while running $ENV{'SCRIPT_NAME'}"
        ),
    ),
    hr(),
    font( 
        { -face => 'Arial' , -size => 4 },
        "Error text: &nbsp;$msg<BR>\n",
        'Please contact the site administrator for help'
    ),
    hr(),
    end_html();

    exit 0 unless ( $cf->{'email_errors'} );

    exit 0 if ( $from_logger );

    if ( ! $cf->{'admin_email'} ) {
        logger("Can't send error email, ADMIN_EMAIL not set!");
        exit 0;
    }
        
    my $subject = "Superfrm error occurred at " .scalar(localtime(time));  
    my $from_header = qq("Superform form processor" <$cf->{'admin_email'}>);
    my $from        = $cf->{'admin_email'};
    my $to          = $cf->{'admin_email'};
    my $text    = <<EOF;

The following error occurred while running $ENV{'SCRIPT_NAME'}:

$msg

To turn off emailed error notification, unset or remove the variable 
EMAIL_ERRORS from the configuration file 

Sincerely,
$0

EOF

    $SENT_ERROR = 1;

    mail_form(\$text, $subject, $from_header, $from, $to);

    exit 0;    
}

#  Log something to log file.

sub logger {
    my $msg = shift;
    return 1 if ( ! $cf->{'logging'} );
    open(LOG,">> $cf->{logfile}") || 
    web_die("Can't write to log file $cf->{logfile}:  $!");
    print LOG time(),"|$msg\n";
    close(LOG);
    debug($msg);
    return 1;
}

#  Check to see if the remote host is allowed to use us; if not,
#  print an error message and exit.

sub check_for_valid_user {

    #  First, make sure we were not called directly.  If so, complain
    #  and exit.

    #  Check referrer .. weak, but keeps malicious newbies out.
    my $referer = 
        (referer() =~ m#^(?:(?:ftp|https?)://)?(.+?)(?:/\S|$)#o)[0];

    my $addr = remote_addr();

    my $host = $ENV{'HTTP_HOST'} || virtual_host() || remote_host();

    if ( $referer =~ m#^\s*$#o ) {
        logger("Warning: $addr called me directly.  Abort.");
        print
        header(),
        start_html(
            '-title' => 'Illegal use of script'
        ),
        center(h1('Sorry, I cannot be called directly')),
        hr(),
        q(
            This script can only be called from an HTML form.&nbsp&nbsp;If
            you believe you have received this message in error, please
            contact the administrator of this site.
        ),
        hr(),
        end_html();

        exit 0;
    }

    #  Check IP address against deny list

    debug("Checking IP address $addr against $cf->{'addr_regex'}");

    if ( $cf->{'addr_regex'} && $addr =~ m/(?:$cf->{'addr_regex'})/oi ) {
        show_deny_page($addr);
        exit 0;
    }

    debug("Checking host $host against $cf->{'allow_regex'}");

    #  Host is OK if it matches our regexp.
    if ( $host !~ m/(?:$cf->{'allow_regex'})/oi ) {
        logger("Host $host denied access");
        show_deny_page($host);
        exit 0;
    }

    debug("Checking referer $referer $addr against $cf->{'allow_regex'}");

    #  Referer is OK if it matches our regexp.
    if ( $referer !~ m/(?:$cf->{'allow_regex'})/oi ) {
        logger("Host $referer denied access");
        show_deny_page($referer);
        exit 0;
    }

    debug("User ($host, $referer, $addr) granted access");

}

sub show_deny_page {

    my $who = shift;

    logger("Access denied for $who");

    print
    header(),
    start_html( { '-title' => "Access denied for $who" } ),
    center( h1("Access denied for $who") ),
    hr(),
    font( 
        { -face => 'Arial' } ,
        qq( 
            Sorry, you are not allowed to use this script. &nbsp;
            If you feel this is an error, please contact the administrator
            of this site.
        ),
        hr(),
        br(), br(),
        b( qq( $who denied access. ) )
    ),
    end_html();

}

#  Check admin values ... retrieve all files specified in cf hash
#  as either regular or HTTP files and replace cf->hash file name
#  with actual contents.

sub check_admin_values {
    my $cf = shift;
    #  These are the admin vars we check for.
    #  s.mailto s.reqfields s.nomail s.noreply );

    #  File related vars ... retrieve content if found.
    my @files = qw ( s.thanks s.mailform s.autoreply s.errorpage );

    my $file = "";  # Hold whole file

    #  If the file has ftp or http, use LWP to get it ... store the file
    #  contents over the variable name.

    #  Write form-keys onto our hash ... only take admin keys, which begin
    #  with `s.'.
    foreach ( (param()) ) {
        next if (! m/^[Ss]\./o or exists $cf->{mf}->{lc($_)});
        $cf->{lc($_)} = param($_);
        debug("cf->admin:  $_ set to " . $cf->{lc($_)});
    }

    for (@files) {

        next if ( ! $cf->{$_} );

        #  If nomail or noreply is set, don't read in files for 
        #  mailform or autoreply.

        next if ($_ eq 's.mailform'  and $cf->{'s.nomail'});
        next if ($_ eq 's.autoreply' and $cf->{'s.noreply'});

        #  MAILTO must be set if MAILFORM is present
        if ( $_ eq 's.mailform' and ! $cf->{'s.mailto'} ) {
            web_die("S.MAILTO must be set if S.MAILFORM is specified");
        }

        #  Looks like our kind of HTTP doc.
        if ( $cf->{"$_"} =~ m/(?:https?|ftp):/o ) {
            debug("Retrieving $_ :  $cf->{$_} via HTTP");
            #  Can we get the file?
            if ( ($file = get($cf->{$_})) eq "" ) {
                logger("Get of $_ ($cf->{$_}) failed");
            }
            #  Set hash value to file contents for later.
            $cf->{$_} = $file;
        } else {

            #  Security!
            #  First, get rid of all sequences of ..[/\] to keep script
            #  kiddies (like myself) out!! :)

            $cf->{$_} =~ s#\.\.[\\/]##go;

            #  Check for chroot command ... if there, substitute beginning
            #  of file with new root ... keep black hats away from passwd
            #  file.

            if ( $cf->{'rootdir'} ne "" ) {
                if ( $^O =~ /Win32/o ) {
                    $cf->{"$_"} =~ s#^(?:[A-Z:\.]*\\)+#$cf->{'rootdir'}#o;
                } else {
                    $cf->{"$_"} =~ s#^(?:\.*/)+#$cf->{'rootdir'}#o;
                }
            }

            debug("Reading $_ :  $cf->{$_} as regular file");

            #  Try as regular file.
            if ( open(F,"< $cf->{$_}") ) {
                local($/) = "";  #  read in one gulp;
                $cf->{$_} = <F>;
                close(F);
            } else {
                logger("Open of $cf->{$_} failed:  $!");
                $cf->{"$_"} = undef;
            }

        }
    }
    
    #  Make sure mailto is set if mailform is specified.
    if ( ! $cf->{'s.nomail'} and $cf->{'s.mailform'} and 
         ! $cf->{'s.mailto'} ) {
         logger("Mailform set but mailto not set: resetting mailform!");
          $cf->{'s.mailform'} = "";
    }

    #  Set default from address if none specified.  we don't validate this
    #  one because it may be faux on purpose and it is NOT passed to the
    #  shell or anything.

    if ( ! $cf->{'s.nomail'} and $cf->{'s.mailfrom'} eq "" ) {
         $cf->{'s.mailfrom'} = qq( <autoreply\@localhost> );
    }

    #  Make sure $cf->{'mailval'}->{'email'} if set if autoreply 
    #  requested.

    if ( ! $cf->{'s.noreply'} and $cf->{'s.autoreply'} and 
         ! $cf->{'mailval'}->{'email'} ) {
        logger("Autoreply set but no return email address set; resetting autoreply");
        $cf->{'s.autoreply'} = "";
    }
  
    #  Do simple check on email addresses to keep bad guys away.
    #  Regexp not great but this should keep out nare-do-wells :)

    if ( ! check_email_addr($cf->{'mailval'}->{'email'}) ) {
        logger("Autoreply address fails check; autoreply NOT sent");
        logger("Passed address was $cf->{'mailval'}->{'email'}");
        $cf->{'s.autoreply'} = "";
        $cf->{'s.noreply'} = 1;
    }

    if ( ! check_email_addr($cf->{'s.mailto'}) ) {
        logger("Email form address fails check; no form will be sent");
        logger("Passed address was $cf->{'s.mailto'}");
        $cf->{'s.mailform'} = "";
        $cf->{'s.nomail'} = 1;
    }

    #  Check CC and BCC addresses

    if ( $cf->{'s.mailcc'} ) {
        validate_and_populate_email_list( \$cf->{'s.mailcc'}, 'Cc' );
    }

    if ( $cf->{'s.mailbcc'} ) {
        validate_and_populate_email_list( \$cf->{'s.mailbcc'}, 'Bcc' );
    }

}

sub validate_and_populate_email_list {
    my $listref = shift;
    my $list_type = shift;  #  In case email validation fails for an address

    debug("Checking $list_type addresses");

    my @temp = split( /\s*,\s*/, $$listref );
    my @dest;

    my $addr;

    for $addr (@temp) {
        if (! check_email_addr($addr)) {
            logger(
              "Email form $list_type address ($addr) fails check; skipping"
            );
        } else {
            debug("Adding $addr to $list_type field");
            push(@dest, $addr);
        }
    }

    #  Save it back into the field as an array ref for later use
    $$listref = \@dest;
}

#  Simple check on email address; we allow single addresses or
#  multiple addresses separated by commas.

sub check_email_addr {
    my $addr = shift;
    my $vchar = '[A-Za-z0-9_!:.-]';

    if ( $addr !~ m/^\s*${vchar}+\@${vchar}+\s*$/o ) {
        return 0;
    }

    return 1;
}

sub read_form_values {
    my $cf = shift;   

    #  What we provide for the user -- so they can use $first etc ...
    my $mt = {
        first    => undef,
        'last'   => undef,
        email    => undef,
        fullname => undef,  
        name     => undef   
    };

    #  Read all non-admin values into hash
    for ( (param()) ) {
        my $nm = lc($_);

        #  Only mail related fields and form fields, so check the
        #  name if it begins with 's.' to make sure it is one we want.
        if ( $_ =~ m#[Ss]\.#o ) {
            next unless (exists $cf->{mf}->{$nm});
        }

        $nm =~ s/^s\.//o;   #  Remove beginning 'S.' part so user doesn't
                            #  have to enter it.

        #  If this is a tag with user's first, last or fullname
        #  or a tag with their email address, add it to mailval hash.
        #  Otherwise put it in form hash.

        if (exists $mt->{$nm} and exists $cf->{'mailval'}->{$nm} ) {
            $cf->{'form'}->{$nm} = $cf->{'mailval'}->{$nm} = param($_);
        } else {
            #  Fix suggested by Steve Perez <steve@technolobo.com>
            #  This will capture single items from a form as well
            #  as mutli-item lists generated by select tags ...
            $cf->{'form'}->{$nm} = join(', ',(param($_)));
        }
        debug("Setting value for $nm to $cf->{'form'}->{$nm}");
    }

    my $fullname = ($cf->{'form'}->{'name'} =~ m/^\s*$/o) ? 
                   $cf->{'form'}->{'fullname'} : $cf->{'form'}->{'name'};

    #  Set full name if not set already and capitalize nicely
    $cf->{'mailval'}->{'first'} = ucfirst(lc($cf->{'mailval'}->{'first'}));
    $cf->{'mailval'}->{'last'}  = ucfirst(lc($cf->{'mailval'}->{'last'}));

    if ( $fullname eq "" ) {
        $fullname = "$cf->{mailval}->{first} $cf->{mailval}->{'last'}";
    } else {
        $fullname = lc($fullname);
    }

    #  Capitalize name properly and add it to form.
    $fullname =~ s/(\b\w)/\U$1\E/go;
    trim(\$fullname);

    $cf->{'form'}->{'fullname'} = $cf->{'mailval'}->{'fullname'} = $fullname;

    #  if first / last names not present, try to get them from full name.
    if ( $cf->{'form'}->{'fullname'} ne "" ) {
        # First
        if ( $cf->{'mailval'}->{'first'} eq "" ) {
            $cf->{'mailval'}->{'first'} = 
                ($fullname =~ m/^([\S\s]+?)(?:\s+\S+)?$/o)[0];
            trim(\$cf->{'mailval'}->{'first'});
        }
        #  Last
        if ( $cf->{'mailval'}->{'last'} eq "" ) {
            $cf->{'mailval'}->{'last'} = ($fullname =~ m/\S+\s+(\S+)$/o)[0];
            trim(\$cf->{'mailval'}->{'last'});
        }
    }

    #  Set these email-related values from CGI env.
    $cf->{'mailval'}->{'remote_host'} = remote_host();
    $cf->{'mailval'}->{'remote_addr'} = remote_addr();
    $cf->{'mailval'}->{'remote_user'} = remote_user();

    if ($cf->{'debug'}) {
        for (keys %{$cf->{'mailval'}} ) {
            debug("Set mailval->$_ to $cf->{'mailval'}->{$_}");
        }
    
        for (keys %{$cf->{'form'}} ) {
            debug("Set form->$_ to $cf->{'form'}->{$_}");
        }
    }
    
}

sub trim {
     $$_[0] =~ s/^\s+//o;
     $$_[0] =~ s/\s+$//o;
}

sub validate_form {
    my $cf = shift;

    @{ $cf->{'badfields'} } = ();

    debug("Validating fields");

    my $nm;
    #  First check for required fields.
    for ( (split(/\s*,\s*/o, $cf->{'s.reqfields'})) ) {

        my ($name,$label) = split(':', $_);

        $label = $name if ! $label;
        $name = lc($name);

        #  If it is an email field, test it properly.
        if ( exists $cf->{mf}->{$name} ) {
            $nm = ( $name =~ m#s\.(.*)$# )[0];
            debug("Checking for EMAIL field $name / $nm");
            if ( $cf->{'mailval'}->{lc($nm)} =~ m/^\s*$/o ) {
                push(@{$cf->{'badfields'}},$label);
                debug("Form missing required email field $nm");
            }
            next;
        }
            
        #  'Normal' form field
        if ( exists $cf->{'form'}->{$name} ) {
            if ( $cf->{'form'}->{$name} =~ m/^\s*$/o ) {
                push(@{$cf->{'badfields'}},$label);
                debug("Form missing required HTML form field $name");
            }
            next;
        }

        #  Administrative field
        if ( exists $cf->{$name} ) {
            if ( $cf->{$name} =~ m/^\s*$/o ) {
                push(@{$cf->{'badfields'}},$label);
                debug("Form missing required admin field $name");
            }
            next;
        }

        web_die(
            qq!
                Field $name does not exist in this form; please check spelling of field 
                names in S.REQFIELDS
            !
        );

    }
    
    #  If any were missing, display an error page and exit.
    if ( ${$cf->{'badfields'}}[0] ne "" ) {
        logger("Required fields missing from form; form not processed");
        display_error_page(\$cf->{'s.errorpage'});
    }
    debug("All fields in form OK.");

}

sub debug {
    my $msg = shift;
    return if ( ! $cf->{debug} );
    open(BUG,">> $cf->{'debugfile'}") || return;
    print BUG scalar(localtime(time)),":  $msg\n";
    close(BUG);
}

sub display_error_page {
    my ($cref) = shift;   #  Fields reference and content reference.

    #  If there is custom content to use, do variable substitution
    #  on it ... otherwise use our own page.

    if ($$cref ne "") {
        substitute_values($cref,ERROR_PAGE);
        print header(), $$cref;
    } else {

        print 
        header(),
        start_html( {'-title' => "Form incomplete"} ),
        center(h1("Form incomplete")),
        hr(),
        font(
            { -face => 'Arial' },
            qq(
                The following required fields were not filled out for 
                this form; please use the <B>back</B> button to return
                to the form and fill them out.
            ),
            br(),
            br(),
            join(', ',map { b($_); } @{ $cf->{'badfields'} }),
            hr(),
            qq(
                Please contact the administrator of this site if you need
                assistance filling out this form.  &nbsp;Thank you.
            )
        ),
        end_html();
    }

    exit 0;
}

sub substitute_values {
    my ($cref, $pg) = @_;
    #  Basic variables available to all pages
    my %vars = (
        fullname => $cf->{'mailval'}->{'fullname'},
        first    => $cf->{'mailval'}->{'first'},
        'last'   => $cf->{'mailval'}->{'last'},
        email    => $cf->{'mailval'}->{'email'},
        remote_host     => $cf->{'mailval'}->{'remote_host'},
        remote_addr     => $cf->{'mailval'}->{'remote_addr'},
        remote_user     => $cf->{'mailval'}->{'remote_user'}
    );

    #  Add all form fields present to the vars hash for use in the template.
    for ( keys %{$cf->{'form'}} ) {
        $vars{"$_"} = $cf->{'form'}->{"$_"};
    }

    #  Add page-specific vars to hash
    if    ( $pg == ERROR_PAGE ) {
        $vars{'field_list'} = join(', ', @{ $cf->{'badfields'} });
    } elsif ( $pg == AUTO_REPLY or $pg == MAIL_FORM or $pg == THANKS_PAGE) {
    } else {
        web_die("substitute_value():  No such page $pg!");
    }

    #  Subsitute special tag $date or ${date} with current date/time
    $$cref =~ s/\${?date}?/scalar(localtime(time))/geo;

    #  Do substitution on content
    #  Vars can be $var, ${var}, or ${var:-NNN}

    for (keys %vars) {
        next if ( $vars{$_} =~ m#^\s*$#o );
        $$cref =~ s/\${?$_(?::-[^}]+)?}?/$vars{"$_"}/g;
    }

    #  Now replace all unsubsituted vars in form ${var:-VAL}
    #  with their default values or blanks if empty.

    $$cref =~ s/\${?\w+(?::-([^}]+))?}?/$1/go;

    return;
}

sub mail_form {
    my ($text_ref, $subject, $from_header, $from, $to, $cc, $bcc) = @_;

    #  Add our mail header to it.

    my $cc_header;
    if ( $cc ) {
       $cc_header = join(', ',@${cc});
    }

    ${$text_ref} = <<EOF;
From:    $from_header
To:      $to
Cc:      $cc_header
Subject: $subject

${$text_ref}

EOF

    #  Mail to TO and CC recipients

    smtp_mail($text_ref, $from, $to, $cc);

    #  Mail to BCC recipients separately

    if ( $bcc ) {
        debug("Bccing form to @{$bcc}");
        my $recipient;
        for $recipient ( @{$bcc} ) {
            smtp_mail($text_ref, $from, $recipient);
        }
    }
}


sub smtp_mail {

    my ($tref, $from, $to, $cc) = @_;

    my $host = "$cf->{mailhost}" || 'localhost';

    debug("Using $host as mail host");

    my $me = Sys::Hostname::hostname();

    if ( ! $me ) {
        web_die("Can't figure out my own hostname:  $!");
    }

    my $smtp = new Net::SMTP("$host", Hello => $me);

    if ( ! $smtp ) {
        logger("SMTP object creation failed: $@");
        web_die("Could not create SMTP object:  $@");
    }

    if (! $smtp->mail($from)) {
        log_smtp_error("SMTP FROM command failed (aborting email)", $smtp);
        $smtp->reset();
        return;
    }

    #
    #  Address message
    #

    #  Send to one or multiple recipients

    if ( ! $cc ) {
        if (! $smtp->to($to)) {
            log_smtp_error("Can't set recipients list to ($to) (no mail sent)", $smtp);
            $smtp->reset();
            return;
        }
    } else {
        if (! $smtp->to($to, @{$cc})) {
            log_smtp_error("Can't set recipients list to ($to @{$cc}) (no mail sent)", $smtp);
            $smtp->reset();
            return;
        }
    }

    select(undef, undef, undef, .75);

    if ( ! $smtp->data() ) {
        log_smtp_error("DATA command rejected", $smtp);
        return;
    }

    for ( (split("\n", $$tref)) ) {
        debug("sending:  $_");
        if (! $smtp->datasend("$_\n")) {
            log_smtp_error("DATASEND failed (aborting email)", $smtp);
            $smtp->reset();
            return;
        }
    }

    if (! $smtp->dataend()) {
        log_smtp_error("Remote server rejected email body", $smtp);
    }

    $smtp->quit();

    debug("Sent email to $to from $from");

}

sub log_smtp_error {

    my $msg  = shift;
    my $smtp = shift;

    my $err = $smtp->code() . ": " . $smtp->message();
    chomp($err);

    logger("$msg: $err");

    return;
}
